#include <stdlib.h>
#include <avr/io.h>

//external variables - declared in global.c
extern unsigned char t_sph;
extern unsigned char t_spl;
extern unsigned char t_r29;
extern unsigned char t_r28;
extern unsigned char task_time;
extern unsigned char ig,jg,kg,lg,mg,ng;
extern unsigned int global_inc_g;
extern unsigned int global_inc;
extern unsigned char temp_g;
extern char lcd_sem;
extern unsigned char lcd_coada[NR_MAX_TASK];


//global variables used only by MyOS functions

//current task index
 char task_curent;

//last task index
 char task_coada;

//first task index
 char task_varf;

//temporal variable
unsigned int tem;

unsigned char  recieve;

task rtos[NR_MAX_TASK];


//functions description

//A. MYOS BASE FUNCTIONS

//init real time kernel
//return - void
//parameters - void
void start_so(void)
{
//init tasks vector
  task_varf=task_coada=task_curent=-1;//
//init semaphore - optional
  lcd_sem = -1;//init LCD semaphore
}

//create a new task
//return - T_EROARE if failed; T_OK if ok
//parameters - void(*taskf) (void) -> pointer to task function (loop). can be any function of type void _name(void)
//           - unsigned char numar -> time period for the task
unsigned char creaza_task(void (*taskf)(void),unsigned char numar)
{
    unsigned int i;

//1.compute task position in list
    if(task_varf==-1)//if the list is empty
    {
      if(task_coada>=NR_MAX_TASK-1) //if the last component index exceed list size
        return T_EROARE; //return error
      task_coada++;//else - increment last component index
      task_curent=task_varf=task_coada;//init first and current component index
    }
    else   //if is at least 1 component in list
    {
      if(task_coada>=NR_MAX_TASK-1) //if the last component exceed list size
        return T_EROARE; //return error
      task_coada++;  //else - increment last component index
    }

    rtos[task_coada].nr_timpi=numar;//init time period for task
    rtos[task_coada].activ=1;//init status on
    for(i=0;i<OSTaskStkSize;i++)
      rtos[task_coada].adr_task[i]=(unsigned char)i;//init task stack
//2.init hardware stack with function pointer
    i = (unsigned int) taskf;
    rtos[task_coada].adr_task[OSTaskStkSizeHard-1]=(unsigned char)i&0x0ff;//lsb
    rtos[task_coada].adr_task[OSTaskStkSizeHard-2]=(unsigned char)((i>>8)&0x0ff);//msb
//init stack pointer
    i = (unsigned int) rtos[task_coada].adr_task+OSTaskStkSizeHard-3;
    rtos[task_coada].adr_task[OSTaskStkSize-1]=(unsigned char)i&0xff;
    rtos[task_coada].adr_task[OSTaskStkSize-2]=(unsigned char)(i>>8)&0xff;
//init Y pointer
    i = (unsigned int) rtos[task_coada].adr_task+62;
    rtos[task_coada].adr_task[OSTaskStkSize-3]=(unsigned char)i&0xff;
    rtos[task_coada].adr_task[OSTaskStkSize-4]=(unsigned char)(i>>8)&0xff;

    return T_OK; //return OK
}

//switch current task from task list
//return - T_EROARE if task switching fail, else T_OK
//parameters - void
unsigned char comuta_task(void)
{
    if(task_varf==-1)//check empty list
      return T_EROARE;
    //  return T_EROARE;

// new current task only from activated tasks
    do
    {
     if(task_curent>=task_coada)//check if current task exceed last task
       task_curent=task_varf-1;
     task_curent++;
     tem++;
    }while(!(rtos[task_curent].activ));

//init current task time
    task_time=rtos[task_curent].nr_timpi;
//init hardware stack
    tem = (unsigned int) ((rtos[task_curent].adr_task)+OSTaskStkSize)>>8;
    t_sph=(unsigned char)(tem&0x00ff);
    t_spl=(unsigned char)((rtos[task_curent].adr_task)+OSTaskStkSize)&0xff;
//init Y pointer
    tem = (unsigned int) ((rtos[task_curent].adr_task)+OSTaskStkSize-2)>>8;
    t_r29=(unsigned char)(tem&0x00ff);
    t_r28=(unsigned char)((rtos[task_curent].adr_task)+OSTaskStkSize-2)&0xff;

    return T_OK;
}


//init t_sp and t_r variables in timer ISR
//return - void
//parameters - void
void get_tsp(void)
{
//init t_sp variables with stack pointer value
    tem = (unsigned int) ((rtos[task_curent].adr_task)+OSTaskStkSize)>>8;
    t_sph=(unsigned char)(tem&0x00ff);
    t_spl=(unsigned char)((rtos[task_curent].adr_task)+OSTaskStkSize)&0xff;
//init t_r variables with Y pointer value
    tem = (unsigned int) ((rtos[task_curent].adr_task)+OSTaskStkSize-2)>>8;
    t_r29=(unsigned char)(tem&0x00ff);
    t_r28=(unsigned char)((rtos[task_curent].adr_task)+OSTaskStkSize-2)&0xff;
}

//init context for first task - used to start multitasking
//return - void
//parameters - void
void init_task(void)
{
 //init t_sp variable with stack pointer
    task_time=rtos[task_curent].nr_timpi;
    tem = (unsigned int) ((rtos[task_curent].adr_task)+OSTaskStkSize)>>8;
    t_sph=(unsigned char)(tem&0x00ff);
    t_spl=(unsigned char)((rtos[task_curent].adr_task)+OSTaskStkSize)&0xff;
//init t_r variables with Y pointer value
    tem = (unsigned int) ((rtos[task_curent].adr_task)+OSTaskStkSize-2)>>8;
    t_r29=(unsigned char)(tem&0x00ff);
    t_r28=(unsigned char)((rtos[task_curent].adr_task)+OSTaskStkSize-2)&0xff;
//init Y pointer to load registers value
    tem = (unsigned int) rtos[task_curent].adr_task+80;
    rtos[task_curent].adr_task[OSTaskStkSize-3]=(unsigned char)tem&0xff;
    rtos[task_curent].adr_task[OSTaskStkSize-4]=(unsigned char)(tem>>8)&0xff;
}

//init stack pointer register and Y pointer with values for first task - used to start multitasking
//return - void
//parameters - void
//NOTE - when call this function because inside modify stack pointer and hardware stack when return automaticaly start first task function!!
void PopSP(void)
{
  cli();//disable interrupt
  asm("  LDS R29,t_sph");//load stack pointer adress in Y pointer registers
  asm("  LDS R28,t_spl");
  asm("  LD R16,-Y");//load stack pointer for first task
  asm("  OUT 0x3d,R16");
  asm("  LD R16,-Y");
  asm("  OUT 0x3e,R16");

//set Y pointer for first task
  asm("  LDS R30,t_r28");
 // asm("  LDS R31,t_0r29");
  asm("  LD R28,-Z");
  asm("  LD R29,-Z");

  b_start_so=1;
 sei();
//when return automaticaly start first task function
}

//save global variables in task stack
//return - void
//parameters - void
//NOTE : if are needed more global variables save here
void push_var(void)
{

  rtos[task_curent].adr_task[OSTaskStkSize-5]=ig;
  rtos[task_curent].adr_task[OSTaskStkSize-6]=jg;
  rtos[task_curent].adr_task[OSTaskStkSize-7]=kg;
  rtos[task_curent].adr_task[OSTaskStkSize-8]=lg;
  rtos[task_curent].adr_task[OSTaskStkSize-9]=mg;
  rtos[task_curent].adr_task[OSTaskStkSize-10]=ng;

//add here another global variable
  rtos[task_curent].adr_task[OSTaskStkSize-11]=temp_g;
 // rtos[task_curent].adr_task[OSTaskStkSize-12]=init_t;
}

//load global variable from task stack
//return - void
//parameters - void
//NOTE : if are needed more global variables load here
void pop_var(void)
{
//unsigned char ig,jg,kg,lg,mg,ng;
  ig=rtos[task_curent].adr_task[OSTaskStkSize-5];
  jg=rtos[task_curent].adr_task[OSTaskStkSize-6];
  kg=rtos[task_curent].adr_task[OSTaskStkSize-7];
  lg=rtos[task_curent].adr_task[OSTaskStkSize-8];
  mg=rtos[task_curent].adr_task[OSTaskStkSize-9];
  ng=rtos[task_curent].adr_task[OSTaskStkSize-10];

  //add here another global variable
  temp_g=rtos[task_curent].adr_task[OSTaskStkSize-11];
 // init_t=rtos[task_curent].adr_task[OSTaskStkSize-12];
}

//SEMAPHORE FUNCTIONS

//context switching when semaphore event occure
//return - void
//parameters - void
//NOTE : this function is similar with timer ISR but this function is manualy called
//       when return restart function for current task
void comuta(void)
{
//disable all interrupts
   cli();

//save registers using Y pointer
  asm("ST      -Y,R27");
  asm("ST      -Y,R26");
  asm("ST      -Y,R25");
  asm("ST      -Y,R24");
  asm("ST      -Y,R31");
  asm("ST      -Y,R30");
  asm("ST      -Y,R3");
  asm("ST      -Y,R2");
  asm("ST      -Y,R1");
  asm("ST      -Y,R0");
  asm("ST      -Y,R23");
  asm("ST      -Y,R22");
  asm("ST      -Y,R21");
  asm("ST      -Y,R20");
  asm("ST      -Y,R19");
  asm("ST      -Y,R18");
  asm("ST      -Y,R17");
  asm("ST      -Y,R16");
  asm("IN      R26,0x3F");
  asm("IN      R27,0x3B");
  asm("  ST -Y, R4");
  asm("  ST -Y, R5");
  asm("  ST -Y, R6");
  asm("  ST -Y, R7");
  asm("  ST -Y, R8");
  asm("  ST -Y, R9");
  asm("  ST -Y, R10");
  asm("  ST -Y, R11");
  asm("  ST -Y, R12");
  asm("  ST -Y, R13");
  asm("  ST -Y, R14");
  asm("  ST -Y, R15");
  asm("  ST -Y, R26");

//init global_inc -
    global_inc=0;
    t_sph=0;
    t_spl=0;
    t_r29=0;
    t_r28=0;
    get_tsp();//init t_sp and t_r variables

//save pointer Y adress - using Z pointer already saved
    asm("  LDS R30,t_r28");//R26 <- t_r28
    asm("  LDS R31,t_r29");//R27 <- t_r29
    asm("  ST -Z, R28");//save Y adress
    asm("  ST -Z, R29");

//save stack pointer
    asm("  LDS R30,t_spl");//save stack pointer using Y pointer already saved
    asm("  LDS R31,t_sph");
    asm("  IN R16,0x3d");//save SPH
    asm("  ST -Z,R16");
    asm("  IN R16,0x3e");//save SPL
    asm("  ST -Z,R16");

    push_var();//save global variables
    comuta_task();//task switching
    pop_var();//restore global variables for new current task

//restore stack pointer register for new curent task
  asm("  LDS R28,t_spl");
  asm("  LDS R29,t_sph");
  asm("  LD R16,-Y");
  asm("  OUT 0x3d,R16");
  asm("  LD R16,-Y");
  asm("  OUT 0x3e,R16");

//restore Y pointer using Z pointer
  asm("  LDS R30,t_r28");
  asm("  LDS R31,t_r29");
  asm("  LD R28,-Z");
  asm("  LD R29,-Z");

//restore all registers
  asm("  LD R26,Y+");
  asm("  LD R15,Y+");
  asm("  LD R14,Y+");
  asm("  LD R13,Y+");
  asm("  LD R12,Y+");
  asm("  LD R11,Y+");
  asm("  LD R10,Y+");
  asm("  LD R9,Y+");
  asm("  LD R8,Y+");
  asm("  LD R7,Y+");
  asm("  LD R6,Y+");
  asm("  LD R5,Y+");
  asm("  LD R4,Y+");

  asm("OUT     0x3B,R27");
  asm("OUT     0x3F,R26");
  asm("LD      R16,Y+");
  asm("LD      R17,Y+");
  asm("LD      R18,Y+");
  asm("LD      R19,Y+");
  asm("LD      R20,Y+");
  asm("LD      R21,Y+");
  asm("LD      R22,Y+");
  asm("LD      R23,Y+");
asm("LD      R0,Y+");
asm("LD      R1,Y+");
asm("LD      R2,Y+");
asm("LD      R3,Y+");
asm("LD      R30,Y+");
asm("LD      R31,Y+");
asm("LD      R24,Y+");
asm("LD      R25,Y+");
asm("LD      R26,Y+");
asm("LD      R27,Y+");

//enable interrupt
  sei();
//when return restart function for current task
}

//request acces to LCD
//return - void
//parameters - void
void lcd_req(void)
{
//test ressource status
  if(lcd_sem>=0)//semaphore busy?
  {
    lcd_coada[lcd_sem]=task_curent;//add curent task in wainting queue for LCD semaphore
    rtos[task_curent].activ = 0;//inactive task
  }
  lcd_sem++;//increment status seamphore
  if(lcd_sem>=NR_MAX_TASK)//test maxim size
    lcd_sem=0;
}

//release lcd resource
//return - void
//parameters - void
void lcd_release(void)
{
 if(lcd_sem>0)//check semaphore status
 {
  rtos[lcd_coada[0]].activ = 1; //reactivate first task in waiting queue for semaphore - this is the first current task
//delete first task from waiting queue
  for(ig=0;ig<lcd_sem-1;ig++)
   lcd_coada[ig]=lcd_coada[ig+1];
  lcd_coada[lcd_sem-1]=0;
 }
 lcd_sem--;//decrement semaphore
 if(lcd_sem<-1)//test minim
  lcd_sem=-1;
}



//3. OTHERS FUNCTIONS

//delay (here ms)
////return - void
//parameters - void
void delay_1_ms(unsigned int time)
{
 unsigned int in;

  in  = global_inc_g;
  for(;;)
  {
   if(in>global_inc_g)
   {
    if(((0xffff-in)+global_inc_g)>=(time*21))
     break;
   }
   else
   {
    if((global_inc_g-in)>=(time*21))
     break;
   }
  }
}

void usart_init(void)
{

}
